//------------------------------------------------------------------------------
// CYmEngine.pas
//------------------------------------------------------------------------------
//
// Description: 	Implementation of the YmPlayer Engine
//
// Autor:			Sledge Hammer of Grinder
//
// Release:		January 2002
//
// This code is based on the AY_EMUL from Sergey Bulba.
// Read YmPlayer.txt in the docs folder for more information.


//------------------------------------------------------------------------------
// Constructor and Destructor
//------------------------------------------------------------------------------

// Constructor
//
constructor CYmEngine.Create;
begin
   // Set internal engine reference
   engine := self;

   // Create the song and the decruncher object
	m_ymSong := CYmSong.Create;
	m_ymDecruncher := CYmDecruncher.Create;

   // Create internal buffers
	New(Buffer1);
	New(Buffer2);

   // Initialize the defaults
	initDefaults;
   m_channelMode := modeManual;
end;

// Destructor
//
destructor CYmEngine.Destroy;
begin
   try
       Stop;
   	Dispose(Buffer1);
		Dispose(Buffer2);
		FreeAndNil(m_ymSong);
		FreeAndNil(m_ymDecruncher);
       engine := nil;
   finally
		inherited;
   end;
end;

//------------------------------------------------------------------------------
// Public declarations
//------------------------------------------------------------------------------

// Loads the ym song from internal ressource
//
function CYmEngine.LoadFromRessource(theFilename: string): boolean;
begin
	// Stop playback
	Stop;

	// Reset the song
	m_ymSong.Reset;

	SetLength(DDrumSamples, 0);

	FileType := m_ymSong.FileType;

	Result := m_ymDecruncher.LoadFromRessource(theFilename);

	if not Result then
		exit;

	m_ymSong.FileName := theFilename;

	// Get the song informations
	getYmInformation;

	m_isLoaded := Result;
end;

// Loads the ym song from file
//
function CYmEngine.LoadFromFile(theFilename: string): boolean;
begin
	// Stop playback
	Stop;

	// Reset the song
	m_ymSong.Reset;

	SetLength(DDrumSamples, 0);

	FileType := m_ymSong.FileType;

	Result := m_ymDecruncher.LoadFromFile(theFilename);

	if not Result then
		exit;

	m_ymSong.FileName := theFilename;

	// Get the song informations
	getYmInformation;

	m_isLoaded := Result;
end;

// Play the song
//
procedure CYmEngine.Play;
var
	Res: Longint;
	ErrMsg: array [0..255] of Char;
	pwfx: pcmwaveformat;
begin
	if not m_isLoaded or m_isPlaying then
		exit;

	pwfx.wf.wFormatTag := 1;
	pwfx.wf.nChannels := NumOfChan;
	pwfx.wf.nSamplesPerSec := SampleRate;
	pwfx.wf.nBlockAlign := (SampleBit div 8) * NumOfChan;
	pwfx.wf.nAvgBytesPerSec := SampleRate * pwfx.wf.nBlockAlign;
	pwfx.wBitsPerSample := SampleBit;
	res := waveOutOpen(@m_phwo, WAVE_MAPPER, @pwfx, DWORD(@waveOutProc), 0,CALLBACK_FUNCTION);
	if res <> 0 then
	begin
		waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
  		raiseWaveOutError(ErrMsg, 'WaveOutOpen')
	end
	else
	begin
		InitForAllTypes;
		m_isPlaying := True;
		MakeBuffer;
		if BuffLen = 0 then
		begin
			res := waveOutClose(m_phwo);
			if res <> 0 then
			begin
				m_isPlaying := False;
               raiseWaveOutError(ErrMsg, 'WaveOutClose(Init)');
			end;
			exit
		end;
		pwh1.lpdata := @Buffer1^;
		pwh1.dwBufferLength := BuffLen * (SampleBit div 8);
		pwh1.dwFlags := 0;
		pwh1.dwUser := 0;
		pwh1.dwLoops := 0;
		res := waveOutPrepareHeader(m_phwo, @pwh1, SizeOf(pwh1));
		if res <> 0 then
		begin
			m_isPlaying := False;
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutPrepareHeaderN1');
			res := waveOutClose(m_phwo);
			if res <> 0 then
			begin
				waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
				raiseWaveOutError(ErrMsg, 'WaveOutClose(Init1)');
			end;
			exit
		end;
		res := waveOutWrite(m_phwo, @pwh1, SizeOf(pwh1));
		if res <> 0 then
		begin
			m_isPlaying := False;
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutWriteN1');
			res := waveOutUnprepareHeader(m_phwo, @pwh1, Sizeof(pwh1));
			if res <> 0 then
			begin
				waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
				raiseWaveOutError(ErrMsg, 'WaveOutUnprepareHeader1(Init)');
			end;
			res := waveOutClose(m_phwo);
			if res <> 0 then
			begin
				waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			    raiseWaveOutError(ErrMsg, 'WaveOutClose(Init2)');
			end;
			exit
		end;
		BaseTime := 0;
		m_isPaused := False;
		if Real_End then
		begin
			pwh2.dwFlags := pwh2.dwFlags or whdr_DONE;
			exit
		end;
		PBuffer := Buffer2;
		MakeBuffer;
		pwh2.lpdata := @Buffer2^;
		pwh2.dwBufferLength := BuffLen * (SampleBit div 8);
		pwh2.dwFlags := 0;
		pwh2.dwUser := 0;
		pwh2.dwLoops := 0;
		if BuffLen = 0 then
		begin
			pwh2.dwFlags := whdr_DONE;
			exit
		end;
		res := waveOutPrepareHeader(m_phwo, @pwh2, SizeOf(pwh2));
		if res <> 0 then
		begin
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutPrepareHeaderN2');
		end;
		res := waveOutWrite(m_phwo, @pwh2, SizeOf(pwh2));
		if res <> 0 then
		begin
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutWriteN2');
		end
	end
end;

// Stop the playback
//
procedure CYmEngine.Stop;
var
	Res: Longint;
	ErrMsg: array [0..255] of Char;
begin
	if m_isPlaying then
	begin
		m_isPlaying := False;
		res := waveOutReset(m_phwo);
		if res <> 0 then
		begin
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutReset');
		end;
		while (pwh1.dwUser = 0) or (pwh2.dwUser = 0) do;
		res := waveOutUnprepareHeader(m_phwo, @pwh1, Sizeof(pwh1));
		if res <> 0 then
		begin
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutUnprepareHeader1(Stop)');
		end;
		res := waveOutUnprepareHeader(m_phwo, @pwh2, Sizeof(pwh2));
		if res <> 0 then
		begin
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutUnprepareHeader2(Stop)');
		end;
		res := waveOutClose(m_phwo);
		if res <> 0 then
		begin
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutClose(Stop)');
		end;
	end
end;

// Pauses the playback
//
procedure CYmEngine.Pause;
var
	Res: Longint;
	ErrMsg: array [0..255] of Char;
begin
	if not m_isPlaying then
		exit;

	if m_isPaused then
	begin
		res := waveOutRestart(m_phwo);
		if res <> 0 then
		begin
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
			raiseWaveOutError(ErrMsg, 'WaveOutRestart');
		end
		else
		begin
			m_isPaused := False;
		end
	end
	else
	begin
		res := waveOutPause(m_phwo);
		if res <> 0 then
		begin
			waveOutGetErrorText(res, ErrMsg, SizeOf(ErrMsg));
		    raiseWaveOutError(ErrMsg, 'WaveOutPause');
		end
		else
		begin
			m_isPaused := True;
		end
	end
end;

// Resets the engine
//
procedure CYmEngine.Reset;
begin
   if m_isPlaying then
   	Stop;

   initDefaults;
end;

// Set the mixer channels
//
procedure CYmEngine.SetMixerModeManual(ALeft, ARight, BLeft, BRight, CLeft, CRight: byte);
begin
	m_channelMode := modeManual;
	Index_AL := ALeft;
	Index_AR := ARight;
	Index_BL := BLeft;
	Index_BR := BRight;
	Index_CL := CLeft;
	Index_CR := CRight;
	calculateLevelTables;
end;

//------------------------------------------------------------------------------
// Private declarations
//------------------------------------------------------------------------------

// Initializes default behaviour
//
procedure CYmEngine.initDefaults;
begin
	initForAllTypes;

	// Set Preamp Volume
	m_preAmp := 255;
	PreAmpMax := m_preAmp;
	SampleRate := SampleRateDef;
	FrqZ80 := FrqZ80Def;
	FrqAY := FrqAYDef;
	MFPTimerMode := MFPTimerModeDef;
	MFPTimerFrq := round(FrqAY * 1.25);
	MaxTStates := MaxTStatesDef;
	Interrupt_Freq := Interrupt_FreqDef;
	Delay_In_Tiks := round(8192 / SampleRate * FrqAY);
	YM6TiksOnInt := FrqAY / (Interrupt_Freq / 1000 * 8);
	Tiks_In_Interrupt := round(FrqAY / (Interrupt_Freq / 1000 * 8));
	YM6SamTiksOnInt := SampleRate / Interrupt_Freq * 1000;
	Sample_Tiks_in_Interrupt := round(SampleRate / Interrupt_Freq * 1000);
	SampleBit := SampleBitDef;
	MaxBuffLen := 65280 shr (SampleBit div 16);
	NumOfChan := NumOfChanDef;
	MaxSamples := MaxBuffLen div NumOfChanDef;
	Index_AL := Index_ALDef;
	Index_AR := Index_ARDef;
	Index_BL := Index_BLDef;
	Index_BR := Index_BRDef;
	Index_CL := Index_CLDef;
	Index_CR := Index_CRDef;
	Emulating_Chip := YM_Chip;
	Synthesizer := Synthesizer_Stereo16;
   setOptimizationFlag(true);
	calculateLevelTables;

	// Reset loading and playing informations
	m_isPlaying := False;
	m_isLoaded := False;
	m_PlayInLoop := False;

   // Reset the ym song
	m_ymSong.Reset;
end;

procedure CYmEngine.getLength;
begin
	Time_Rasch := 0;
	VBL_For_Loop := 0;
	Global_Tick_Max := Number_Of_Tiks_In_VTX;
	Time_Rasch := Round(Number_Of_Tiks_In_VTX / Interrupt_Freq * 1000000);
end;

procedure CYmEngine.synthesizerZX50;
begin
	if not IntFlag then
	begin
		if m_optimizationForQuality then
			Number_Of_Tiks.hi := Tiks_In_Interrupt
		else
			Number_Of_Tiks.hi := Sample_Tiks_In_Interrupt
	end
	else
		IntFlag := False;
	Synthesizer
end;

procedure CYmEngine.makeBufferVTX;
begin
	Bufflen := 0;
	if IntFlag then synthesizerZX50;
	if IntFlag then exit;
	if Global_Tick_Counter >= Global_Tick_Max then
		if m_PlayInLoop then
			Global_Tick_Counter := Global_Tick_Max
		else
		begin
			Real_End := True;
			exit;
		end;
	repeat
		vtxYm3bGetRegisters;
		synthesizerZX50;
		if (Global_Tick_Counter >= Global_Tick_Max) and (not IntFlag) then
			if m_PlayInLoop then
				Global_Tick_Counter := Global_Tick_Max
			else
			begin
				Real_End := True;
				exit;
			end;
		if Position_In_Vtx = Number_of_tiks_in_VTX then Position_In_Vtx := VBL_For_Loop;
	until Real_End or (BuffLen = MaxBuffLen);
end;

procedure CYmEngine.vtxYm3bGetRegisters;
var
	i: word;
	k: longword;
begin
	k := VTX_Offset;
	for i := 0 to 12 do
	begin
		case i of
			1,3,5:
				RegisterAY.Index[i] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 15;
			6: RegisterAY.Noise := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 31;
			7: SetMixerRegister(m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 63);
			8: SetAmplA(m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 31);
			9: SetAmplB(m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 31);
			10: SetAmplC(m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 31);
			else
				RegisterAY.Index[i] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
		end;
		inc(k, Number_of_tiks_in_VTX);
	end;
	if m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] <> 255 then
		SetEnvelopeRegister(m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 15);
	inc(Global_Tick_Counter);
	inc(Position_In_VTX)
end;

procedure CYmEngine.makeBufferYM6;
var
	MaxT: real;
begin
	Bufflen := 0;
	if IntFlag then synthesizerYM6;
	if IntFlag then exit;
	if Global_Tick_Counter >= Global_Tick_Max then
		if m_PlayInLoop then
			Global_Tick_Counter := Global_Tick_Max
		else
		begin
			Real_End := True;
			exit
		end;
	if m_optimizationForQuality then
		MaxT := YM6TiksOnInt
	else
		MaxT := YM6SamTiksOnInt;
	if m_ymDecruncher.DecrunchedData^[19] and 1 <> 0 then
	begin
		if Position_In_Vtx = Number_of_tiks_in_VTX then
			Position_In_Vtx := VBL_For_Loop;
		repeat
			if YM6CurTik >= MaxT then
			begin
				YM6CurTik := YM6CurTik - MaxT;
				ym6iGetRegisters
			end;
			ym6GetExtraRegisters;
			synthesizerYM6;
			if (Global_Tick_Counter >= Global_Tick_Max) and (not IntFlag) then
				if m_PlayInLoop then Global_Tick_Counter := Global_Tick_Max
				else
				begin
					Real_End := True;
					exit
				end;
			if Position_In_Vtx = Number_of_tiks_in_VTX then
				Position_In_Vtx := VBL_For_Loop;
		until Real_End or (BuffLen = MaxBuffLen);
	end
	else
		repeat
			if YM6CurTik >= MaxT then
			begin
				YM6CurTik := YM6CurTik - MaxT;
				ym6GetRegisters
			end;
			ym6GetExtraRegisters;
			synthesizerYM6;
			if (Global_Tick_Counter >= Global_Tick_Max) and (not IntFlag) then
				if m_PlayInLoop then
					Global_Tick_Counter := Global_Tick_Max
				else
				begin
					Real_End := True;
					exit
				end;
			if (Position_In_Vtx div 16 = Number_of_tiks_in_VTX) then
				Position_In_Vtx := VBL_For_Loop * 16;
		until Real_End or (BuffLen = MaxBuffLen)
	end;

procedure CYmEngine.ym6GetExtraRegisters;
var
	t1, t2, t3: real;
begin
	if m_optimizationForQuality then
		t3 := YM6TiksOnInt - YM6CurTik
	else
		t3 := YM6SamTiksOnInt - YM6CurTik;
	t1 := t3;
	t2 := t3;
	if AtariSE1Channel <> 0 then
	begin
		if AtariTimerCounter1 = 0 then
			case AtariSE1Type of
				0:
				begin
					if AtariV1 = 0 then
						AtariV1 := AtariParam1
					else
						AtariV1 := 0;
					RegisterAY.Index[7 + AtariSE1Channel] := AtariV1
				end;
				1:
				begin
					RegisterAY.Index[7 + AtariSE1Channel] :=
						DDrumSamples[AtariParam1].Buf[AtariSE1Pos];
					inc(AtariSE1Pos);
					if AtariSE1Pos >= DDrumSamples[AtariParam1].Length then
						AtariSE1Channel := 0
				end;
				2:
				begin
					RegisterAY.Index[7 + AtariSE1Channel] :=
						YM6SinusTable[AtariParam1, YM6SinusPos1];
					YM6SinusPos1 := (YM6SinusPos1 + 1) and 7
				end;
				3: SetEnvelopeRegister(AtariParam1);
			end;
		t1 := AtariTimerPeriod1 - AtariTimerCounter1
	end;
	if AtariSE2Channel <> 0 then
	begin
		if AtariTimerCounter2 = 0 then
			case AtariSE2Type of
				0:
				begin
					if AtariV2 = 0 then
						AtariV2 := AtariParam2
					else
						AtariV2 := 0;
					RegisterAY.Index[7 + AtariSE2Channel] := AtariV2
				end;
				1:
				begin
					RegisterAY.Index[7 + AtariSE2Channel] :=
						DDrumSamples[AtariParam2].Buf[AtariSE2Pos] and 15;
					inc(AtariSE2Pos);
					if AtariSE2Pos >= DDrumSamples[AtariParam2].Length then
						AtariSE2Channel := 0
				end;
				2:
				begin
					RegisterAY.Index[7 + AtariSE2Channel] :=
						YM6SinusTable[AtariParam2, YM6SinusPos2];
					YM6SinusPos2 := (YM6SinusPos2 + 1) and 7
				end;
				3: SetEnvelopeRegister(AtariParam2);
			end;
		t2 := AtariTimerPeriod2 - AtariTimerCounter2
	end;

	if t2 < t1 then t1 := t2;
	if t3 < t1 then t1 := t3;

	YM6Tiks := round(t1 * 4294967296);

	if AtariSE1Channel <> 0 then
	begin
		AtariTimerCounter1 := AtariTimerCounter1 + t1;
		if AtariTimerCounter1 >= AtariTimerPeriod1 then
			AtariTimerCounter1 := 0
	end;
	if AtariSE2Channel <> 0 then
	begin
		AtariTimerCounter2 := AtariTimerCounter2 + t1;
		if AtariTimerCounter2 >= AtariTimerPeriod2 then
			AtariTimerCounter2 := 0
	end;
	YM6CurTik := YM6CurTik + t1
end;

procedure CYmEngine.makeBufferYM5;
var
	MaxT: real;
begin
	Bufflen := 0;
	if IntFlag then synthesizerYM6;
	if IntFlag then exit;
	if Global_Tick_Counter >= Global_Tick_Max then
		if m_PlayInLoop then
			Global_Tick_Counter := Global_Tick_Max
		else
		begin
			Real_End := True;
			exit
		end;
	if m_optimizationForQuality then
		MaxT := YM6TiksOnInt
	else
		MaxT := YM6SamTiksOnInt;
	if m_ymDecruncher.DecrunchedData^[19] and 1 <> 0 then
	begin
		if Position_In_Vtx = Number_of_tiks_in_VTX then
			Position_In_Vtx := VBL_For_Loop;
		repeat
			if YM6CurTik >= MaxT then
			begin
				YM6CurTik := YM6CurTik - MaxT;
				ym5iGetRegisters
			end;
			ym6GetExtraRegisters;
			synthesizerYM6;
			if (Global_Tick_Counter >= Global_Tick_Max) and (not IntFlag) then
				if m_PlayInLoop then
					Global_Tick_Counter := Global_Tick_Max
				else
				begin
					Real_End := True;
					exit
				end;
			if Position_In_Vtx = Number_of_tiks_in_VTX then
				Position_In_Vtx := VBL_For_Loop;
		until Real_End or (BuffLen = MaxBuffLen)
		end
		else
			repeat
				if YM6CurTik >= MaxT then
			begin
				YM6CurTik := YM6CurTik - MaxT;
				ym5GetRegisters
			end;
			ym6GetExtraRegisters;
			synthesizerYM6;
			if (Global_Tick_Counter >= Global_Tick_Max) and (not IntFlag) then
				if m_PlayInLoop then
					Global_Tick_Counter := Global_Tick_Max
				else
				begin
					Real_End := True;
					exit
				end;
			if (Position_In_Vtx div 16 = Number_of_tiks_in_VTX) then
				Position_In_Vtx := VBL_For_Loop * 16;
		until Real_End or (BuffLen = MaxBuffLen)
	end;

procedure CYmEngine.ym6iGetRegisters;
var
	k: longword;
	b: byte;
	SE1Ch, SE2Ch, SE1Typ, Se2Typ, SE1TC, SE2TC: byte;
	frq: real;
begin
	k := VTX_Offset;
	RegisterAY.Index[0] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Index[1] := b and 15;
	SE1Ch := b and $30 shr 4;
	SE1Typ := b shr 6;

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[2] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Index[3] := b and 15;
	SE2Ch := b and $30 shr 4;
	SE2Typ := b shr 6;

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[4] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[5] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 15;

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Noise := b and 31;
	if SE1Ch <> 0 then
		AtariSE1TP := b shr 5;

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 63;
	if SE1Typ = 1 then
		case SE1Ch of
			1: b := b or 9;
			2: b := b or 18;
			3: b := b or 36
		end;
	if SE2Typ = 1 then
		case SE2Ch of
			1: b := b or 9;
			2: b := b or 18;
			3: b := b or 36
		end;
	if AtariSE1Type = 1 then
		case AtariSE1Channel of
			1: if b and 9 <> 9 then AtariSE1Channel := 0;
			2: if b and 18 <> 18 then AtariSE1Channel := 0;
			3: if b and 36 <> 36 then AtariSE1Channel := 0;
		end;
	if AtariSE2Type = 1 then
		case AtariSE2Channel of
			1: if b and 9 <> 9 then AtariSE2Channel := 0;
			2: if b and 18 <> 18 then AtariSE2Channel := 0;
			3: if b and 36 <> 36 then AtariSE2Channel := 0;
		end;
	SetMixerRegister(b);

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if SE2Ch <> 0 then
		AtariSE2TP := b shr 5;
	if SE1Ch = 1 then
	begin
		if SE1Typ in [0,2] then
		begin
			AtariParam1 := b and 15;
			Envelope_EnA := True
		end
		else if SE1Typ = 3 then
		begin
			AtariParam1 := b and 15;
			SetAmplA(b and 16)
		end
		else
		begin
			AtariParam1 := b and 31;
			Envelope_EnA := True
		end
	end
	else if SE2Ch = 1 then
	begin
		if SE2Typ in [0,2] then
		begin
			AtariParam2 := b and 15;
			Envelope_EnA := True
		end
		else if SE2Typ = 3 then
		begin
			AtariParam2 := b and 15;
			SetAmplA(b and 16)
		end
		else
		begin
			AtariParam2 := b and 31;
			Envelope_EnA := True
		end
	end
	else if not ((AtariSE1Channel = 1) and (AtariSE1Type = 1)) and
		not ((AtariSE2Channel = 1) and (AtariSE2Type = 1)) then
		SetAmplA(b and 31);

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if SE1Ch = 2 then
	begin
		if SE1Typ in [0,2] then
		begin
			AtariParam1 := b and 15;
			Envelope_EnB := True
		end
		else if SE1Typ = 3 then
		begin
			AtariParam1 := b and 15;
			SetAmplB(b and 16)
		end
		else
		begin
			AtariParam1 := b and 31;
			Envelope_EnB := True
		end
	end
	else if SE2Ch = 2 then
	begin
		if SE2Typ in [0,2] then
		begin
			AtariParam2 := b and 15;
			Envelope_EnB := True
		end
		else if SE2Typ = 3 then
		begin
			AtariParam2 := b and 15;
			SetAmplB(b and 16)
		end
		else
		begin
			AtariParam2 := b and 31;
			Envelope_EnB := True
		end
	end
	else if not ((AtariSE1Channel = 2) and (AtariSE1Type = 1)) and
		not ((AtariSE2Channel = 2) and (AtariSE2Type = 1)) then
		SetAmplB(b and 31);

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if SE1Ch = 3 then
	begin
		if SE1Typ in [0,2] then
		begin
			AtariParam1 := b and 15;
			Envelope_EnC := True
		end
		else if SE1Typ = 3 then
		begin
			AtariParam1 := b and 15;
			SetAmplC(b and 16)
		end
		else
		begin
			AtariParam1 := b and 31;
			Envelope_EnC := True
		end
	end
	else if SE2Ch = 3 then
	begin
		if SE2Typ in [0,2] then
		begin
			AtariParam2 := b and 15;
			Envelope_EnC := True
		end
		else if SE2Typ = 3 then
		begin
			AtariParam2 := b and 15;
			SetAmplC(b and 16)
		end
		else
		begin
			AtariParam2 := b and 31;
			Envelope_EnC := True
		end
	end
	else if not ((AtariSE1Channel = 3) and (AtariSE1Type = 1)) and
		not ((AtariSE2Channel = 3) and (AtariSE2Type = 1)) then
		SetAmplC(b and 31);

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[11] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[12] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if b <> 255 then
		SetEnvelopeRegister(b and 15);

	inc(k, Number_of_tiks_in_VTX);
	SE1TC := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	SE2TC := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	if SE1Ch <> 0 then
	begin
		AtariSE1Type := SE1Typ;
		AtariSE1Channel := SE1Ch;
		AtariSE1Pos := 0;
		if SE1TC <> 0 then
		begin
			if m_optimizationForQuality then
				frq := 1 / (MFPTimerFrq / SE1TC / (FrqAY / 8))
			else
				frq := 1 / (MFPTimerFrq / SE1TC / SampleRate);
			case AtariSE1TP of
				0: AtariSE1Channel := 0;
				1: AtariTimerPeriod1 := frq * 4;
				2: AtariTimerPeriod1 := frq * 10;
				3: AtariTimerPeriod1 := frq * 16;
				4: AtariTimerPeriod1 := frq * 50;
				5: AtariTimerPeriod1 := frq * 64;
				6: AtariTimerPeriod1 := frq * 100;
				7: AtariTimerPeriod1 := frq * 200
			end
		end
		else
			AtariSE1Channel := 0;
		if AtariTimerCounter1 >= AtariTimerPeriod1 then
			AtariTimerCounter1 := 0
	end
	else
	begin
		if AtariSE1Type <> 1 then
			AtariSE1Channel := 0;
		AtariTimerCounter1 := 0;
		AtariV1 := 0
	end;
	if SE2Ch <> 0 then
	begin
		AtariSE2Type := SE2Typ;
		AtariSE2Channel := SE2Ch;
		AtariSE2Pos := 0;
		if SE2TC <> 0 then
		begin
			if m_optimizationForQuality then
				frq := 1 / (MFPTimerFrq / SE2TC / (FrqAY / 8))
			else
				frq := 1 / (MFPTimerFrq / SE2TC / SampleRate);
			case AtariSE2TP of
				0: AtariSE2Channel := 0;
				1: AtariTimerPeriod2 := frq * 4;
				2: AtariTimerPeriod2 := frq * 10;
				3: AtariTimerPeriod2 := frq * 16;
				4: AtariTimerPeriod2 := frq * 50;
				5: AtariTimerPeriod2 := frq * 64;
				6: AtariTimerPeriod2 := frq * 100;
				7: AtariTimerPeriod2 := frq * 200
			end
		end
		else
			AtariSE2Channel := 0;
		if AtariTimerCounter2 >= AtariTimerPeriod2 then
			AtariTimerCounter2 := 0
	end
	else
	begin
		if AtariSE2Type <> 1 then
			AtariSE2Channel := 0;
		AtariTimerCounter2 := 0;
		AtariV2 := 0
	end;

	inc(Global_Tick_Counter);
	inc(Position_In_VTX)
end;

procedure CYmEngine.ym6GetRegisters;
var
	k: longword;
	b: byte;
	SE1Ch, SE2Ch, SE1Typ, Se2Typ, SE1TC, SE2TC: byte;
	frq: real;
begin
	k := VTX_Offset;
	RegisterAY.Index[0] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Index[1] := b and 15;
	SE1Ch := b and $30 shr 4;
	SE1Typ := b shr 6;

	inc(k);
	RegisterAY.Index[2] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Index[3] := b and 15;
	SE2Ch := b and $30 shr 4;
	SE2Typ := b shr 6;

	inc(k);
	RegisterAY.Index[4] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	RegisterAY.Index[5] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 15;

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Noise := b and 31;
	if SE1Ch <> 0 then
		AtariSE1TP := b shr 5;

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 63;
	if SE1Typ = 1 then
		case SE1Ch of
			1: b := b or 9;
			2: b := b or 18;
			3: b := b or 36
		end;
	if SE2Typ = 1 then
		case SE2Ch of
			1: b := b or 9;
			2: b := b or 18;
			3: b := b or 36
		end;
	if AtariSE1Type = 1 then
		case AtariSE1Channel of
			1: if b and 9 <> 9 then AtariSE1Channel := 0;
			2: if b and 18 <> 18 then AtariSE1Channel := 0;
			3: if b and 36 <> 36 then AtariSE1Channel := 0;
		end;
	if AtariSE2Type = 1 then
		case AtariSE2Channel of
			1: if b and 9 <> 9 then AtariSE2Channel := 0;
			2: if b and 18 <> 18 then AtariSE2Channel := 0;
			3: if b and 36 <> 36 then AtariSE2Channel := 0;
		end;
	SetMixerRegister(b);

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if SE2Ch <> 0 then
		AtariSE2TP := b shr 5;
	if SE1Ch = 1 then
	begin
		if SE1Typ in [0,2] then
		begin
			AtariParam1 := b and 15;
			Envelope_EnA := True
		end
		else if SE1Typ = 3 then
		begin
			AtariParam1 := b and 15;
			SetAmplA(b and 16)
		end
		else
		begin
			AtariParam1 := b and 31;
			Envelope_EnA := True
		end
	end
	else if SE2Ch = 1 then
	begin
		if SE2Typ in [0,2] then
		begin
			AtariParam2 := b and 15;
			Envelope_EnA := True
		end
		else if SE2Typ = 3 then
		begin
			AtariParam2 := b and 15;
			SetAmplA(b and 16)
		end
		else
		begin
			AtariParam2 := b and 31;
			Envelope_EnA := True
		end
	end
	else if not ((AtariSE1Channel = 1) and (AtariSE1Type = 1)) and
		not ((AtariSE2Channel = 1) and (AtariSE2Type = 1)) then
		SetAmplA(b and 31);

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if SE1Ch = 2 then
	begin
		if SE1Typ in [0,2] then
		begin
			AtariParam1 := b and 15;
			Envelope_EnB := True
		end
		else if SE1Typ = 3 then
		begin
			AtariParam1 := b and 15;
			SetAmplB(b and 16)
		end
		else
		begin
			AtariParam1 := b and 31;
			Envelope_EnB := True
		end
	end
	else if SE2Ch = 2 then
	begin
		if SE2Typ in [0,2] then
		begin
			AtariParam2 := b and 15;
			Envelope_EnB := True
		end
		else if SE2Typ = 3 then
		begin
			AtariParam2 := b and 15;
			SetAmplB(b and 16)
		end
		else
		begin
			AtariParam2 := b and 31;
			Envelope_EnB := True
		end
	end
	else if not ((AtariSE1Channel = 2) and (AtariSE1Type = 1)) and
		not ((AtariSE2Channel = 2) and (AtariSE2Type = 1)) then
		SetAmplB(b and 31);

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if SE1Ch = 3 then
	begin
		if SE1Typ in [0,2] then
		begin
			AtariParam1 := b and 15;
			Envelope_EnC := True
		end
		else if SE1Typ = 3 then
		begin
			AtariParam1 := b and 15;
			SetAmplC(b and 16)
		end
		else
		begin
			AtariParam1 := b and 31;
			Envelope_EnC := True
		end
	end
	else if SE2Ch = 3 then
	begin
		if SE2Typ in [0,2] then
		begin
			AtariParam2 := b and 15;
			Envelope_EnC := True
		end
		else if SE2Typ = 3 then
		begin
			AtariParam2 := b and 15;
			SetAmplC(b and 16)
		end
		else
		begin
			AtariParam2 := b and 31;
			Envelope_EnC := True
		end
	end
	else if not ((AtariSE1Channel = 3) and (AtariSE1Type = 1)) and
		not ((AtariSE2Channel = 3) and (AtariSE2Type = 1)) then
		SetAmplC(b and 16);

	inc(k);
	RegisterAY.Index[11] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	RegisterAY.Index[12] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if b <> 255 then
		SetEnvelopeRegister(b and 15);

	inc(k);
	SE1TC := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	SE2TC := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	if SE1Ch <> 0 then
	begin
		AtariSE1Type := SE1Typ;
		AtariSE1Channel := SE1Ch;
		AtariSE1Pos := 0;
		if SE1TC <> 0 then
		begin
			if m_optimizationForQuality then
				frq := 1 / (MFPTimerFrq / SE1TC / (FrqAY / 8))
			else
				frq := 1 / (MFPTimerFrq / SE1TC / SampleRate);
			case AtariSE1TP of
				0: AtariSE1Channel := 0;
				1: AtariTimerPeriod1 := frq * 4;
				2: AtariTimerPeriod1 := frq * 10;
				3: AtariTimerPeriod1 := frq * 16;
				4: AtariTimerPeriod1 := frq * 50;
				5: AtariTimerPeriod1 := frq * 64;
				6: AtariTimerPeriod1 := frq * 100;
				7: AtariTimerPeriod1 := frq * 200
			end
		end
		else
			AtariSE1Channel := 0;
		if AtariTimerCounter1 >= AtariTimerPeriod1 then
			AtariTimerCounter1 := 0
	end
	else
	begin
		if AtariSE1Type <> 1 then
			AtariSE1Channel := 0;
		AtariTimerCounter1 := 0;
		AtariV1 := 0
	end;
	if SE2Ch <> 0 then
	begin
		AtariSE2Type := SE2Typ;
		AtariSE2Channel := SE2Ch;
		AtariSE2Pos := 0;
		if SE2TC <> 0 then
		begin
			if m_optimizationForQuality then
				frq := 1 / (MFPTimerFrq / SE2TC / (FrqAY / 8))
			else
				frq := 1 / (MFPTimerFrq / SE2TC / SampleRate);
			case AtariSE2TP of
				0: AtariSE2Channel := 0;
				1: AtariTimerPeriod2 := frq * 4;
				2: AtariTimerPeriod2 := frq * 10;
				3: AtariTimerPeriod2 := frq * 16;
				4: AtariTimerPeriod2 := frq * 50;
				5: AtariTimerPeriod2 := frq * 64;
				6: AtariTimerPeriod2 := frq * 100;
				7: AtariTimerPeriod2 := frq * 200
			end
		end
		else
			AtariSE2Channel := 0;
		if AtariTimerCounter2 >= AtariTimerPeriod2 then
			AtariTimerCounter2 := 0
	end
	else
	begin
		if AtariSE2Type <> 1 then
			AtariSE2Channel := 0;
		AtariTimerCounter2 := 0;
		AtariV2 := 0
	end;

	inc(Global_Tick_Counter);
	inc(Position_In_VTX, 16)
end;

procedure CYmEngine.ym5iGetRegisters;
var
	k: longword;
	b: byte;
	DD, SE1TC, SE2TC: byte;
	frq: real;
begin
	k := VTX_Offset;
	RegisterAY.Index[0] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Index[1] := b and 15;
	AtariSE1Channel := b and $30 shr 4;
	AtariSE1Type := 0;
	if (AtariSE1Channel <> 0) and (b and $40 <> 0) then
		AtariTimerCounter1 := 0;

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[2] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Index[3] := b and 15;
	DD := b and $30 shr 4;
	AtariSE2Type := 1;

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[4] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[5] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 15;

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Noise := b and 31;
	if AtariSE1Channel <> 0 then
		AtariSE1TP := b shr 5;

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 63;
	case DD of
		0: case AtariSE2Channel of
				1: if b and 9 <> 9 then AtariSE2Channel := 0;
				2: if b and 18 <> 18 then AtariSE2Channel := 0;
				3: if b and 36 <> 36 then AtariSE2Channel := 0;
			end;
		1: b := b or 9;
		2: b := b or 18;
		3: b := b or 36;
	end;
	SetMixerRegister(b);

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if DD <> 0 then
		AtariSE2TP := b shr 5;
	if AtariSE1Channel = 1 then
	begin
		AtariParam1 := b and 15;
		Envelope_EnA := True
	end
	else if DD = 1 then
	begin
		AtariParam2 := b and 31;
		Envelope_EnA := True
	end
	else if AtariSE2Channel <> 1 then
		SetAmplA(b and 31);

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if AtariSE1Channel = 2 then
	begin
		AtariParam1 := b and 15;
		Envelope_EnB := True
	end
	else if DD = 2 then
	begin
		AtariParam2 := b and 31;
		Envelope_EnB := True
	end
	else if AtariSE2Channel <> 2 then
		SetAmplB(b and 31);

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if AtariSE1Channel = 3 then
	begin
		AtariParam1 := b and 15;
		Envelope_EnC := True
	end
	else if DD = 3 then
	begin
		AtariParam2 := b and 31;
		Envelope_EnC := True
	end
	else if AtariSE2Channel <> 3 then
		SetAmplC(b and 31);

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[11] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	RegisterAY.Index[12] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if b <> 255 then
		SetEnvelopeRegister(b and 15);

	inc(k, Number_of_tiks_in_VTX);
	SE1TC := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k, Number_of_tiks_in_VTX);
	SE2TC := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	if AtariSE1Channel <> 0 then
	begin
		if SE1TC <> 0 then
		begin
			if m_optimizationForQuality then
				frq := 1 / (MFPTimerFrq / SE1TC / (FrqAY / 8))
			else
				frq := 1 / (MFPTimerFrq / SE1TC / SampleRate);
			case AtariSE1TP of
				0: AtariSE1Channel := 0;
				1: AtariTimerPeriod1 := frq * 4;
				2: AtariTimerPeriod1 := frq * 10;
				3: AtariTimerPeriod1 := frq * 16;
				4: AtariTimerPeriod1 := frq * 50;
				5: AtariTimerPeriod1 := frq * 64;
				6: AtariTimerPeriod1 := frq * 100;
				7: AtariTimerPeriod1 := frq * 200
			end
		end
		else
			AtariSE1Channel := 0;
		if AtariTimerCounter1 >= AtariTimerPeriod1 then
			AtariTimerCounter1 := 0
	end
	else
	begin
		AtariTimerCounter1 := 0;
		AtariV1 := 0
	end;

	if DD <> 0 then
	begin
		AtariSE2Channel := DD;
		AtariSE2Pos := 0;
		if SE2TC <> 0 then
		begin
			if m_optimizationForQuality then
				frq := 1 / (MFPTimerFrq / SE2TC / (FrqAY / 8))
			else
				frq := 1 / (MFPTimerFrq / SE2TC / SampleRate);
			case AtariSE2TP of
				0: AtariSE2Channel := 0;
				1: AtariTimerPeriod2 := frq * 4;
				2: AtariTimerPeriod2 := frq * 10;
				3: AtariTimerPeriod2 := frq * 16;
				4: AtariTimerPeriod2 := frq * 50;
				5: AtariTimerPeriod2 := frq * 64;
				6: AtariTimerPeriod2 := frq * 100;
				7: AtariTimerPeriod2 := frq * 200
			end
		end
		else
			AtariSE2Channel := 0;
		if AtariTimerCounter2 >= AtariTimerPeriod2 then
			AtariTimerCounter2 := 0
	end
	else
		AtariTimerCounter2 := 0;

	inc(Global_Tick_Counter);
	inc(Position_In_VTX)
end;

procedure CYmEngine.ym5GetRegisters;
var
	k: longword;
	b: byte;
	DD, SE1TC, SE2TC: byte;
	frq: real;
begin
	k := VTX_Offset;
	RegisterAY.Index[0] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Index[1] := b and 15;
	AtariSE1Channel := b and $30 shr 4;
	AtariSE1Type := 0;
	if (AtariSE1Channel <> 0) and (b and $40 <> 0) then
		AtariTimerCounter1 := 0;

	inc(k);
	RegisterAY.Index[2] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Index[3] := b and 15;
	DD := b and $30 shr 4;
	AtariSE2Type := 1;

	inc(k);
	RegisterAY.Index[4] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	RegisterAY.Index[5] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 15;

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	RegisterAY.Noise := b and 31;
	if AtariSE1Channel <> 0 then
		AtariSE1TP := b shr 5;

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 63;
	case DD of
		0: case AtariSE2Channel of
				1: if b and 9 <> 9 then AtariSE2Channel := 0;
				2: if b and 18 <> 18 then AtariSE2Channel := 0;
				3: if b and 36 <> 36 then AtariSE2Channel := 0;
			end;
		1: b := b or 9;
		2: b := b or 18;
		3: b := b or 36;
	end;
	SetMixerRegister(b);

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if DD <> 0 then
		AtariSE2TP := b shr 5;
	if AtariSE1Channel = 1 then
	begin
		AtariParam1 := b and 15;
		Envelope_EnA := True
	end
	else if DD = 1 then
	begin
		AtariParam2 := b and 31;
		Envelope_EnA := True
	end
	else if AtariSE2Channel <> 1 then
		SetAmplA(b and 31);

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if AtariSE1Channel = 2 then
	begin
		AtariParam1 := b and 15;
		Envelope_EnB := True
	end
	else if DD = 2 then
	begin
		AtariParam2 := b and 31;
		Envelope_EnB := True
	end
	else if AtariSE2Channel <> 2 then
		SetAmplB(b and 31);

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k] and 31;
	if AtariSE1Channel = 3 then
	begin
		AtariParam1 := b and 15;
		Envelope_EnC := True
	end
	else if DD = 3 then
	begin
		AtariParam2 := b and 31;
		Envelope_EnC := True
	end
	else if AtariSE2Channel <> 3 then
		SetAmplC(b and 31);

	inc(k);
	RegisterAY.Index[11] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	RegisterAY.Index[12] := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	b := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];
	if b <> 255 then
		SetEnvelopeRegister(b and 15);

	inc(k);
	SE1TC := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	inc(k);
	SE2TC := m_ymDecruncher.DecrunchedData^[Position_In_Vtx + k];

	if AtariSE1Channel <> 0 then
	begin
		if SE1TC <> 0 then
		begin
			if m_optimizationForQuality then
				frq := 1 / (MFPTimerFrq / SE1TC / (FrqAY / 8))
			else
				frq := 1 / (MFPTimerFrq / SE1TC / SampleRate);
			case AtariSE1TP of
				0: AtariSE1Channel := 0;
				1: AtariTimerPeriod1 := frq * 4;
				2: AtariTimerPeriod1 := frq * 10;
				3: AtariTimerPeriod1 := frq * 16;
				4: AtariTimerPeriod1 := frq * 50;
				5: AtariTimerPeriod1 := frq * 64;
				6: AtariTimerPeriod1 := frq * 100;
				7: AtariTimerPeriod1 := frq * 200
			end
		end
		else
			AtariSE1Channel := 0;
		if AtariTimerCounter1 >= AtariTimerPeriod1 then
			AtariTimerCounter1 := 0
	end
	else
	begin
		AtariTimerCounter1 := 0;
		AtariV1 := 0
	end;

	if DD <> 0 then
	begin
		AtariSE2Channel := DD;
		AtariSE2Pos := 0;
		if SE2TC <> 0 then
		begin
			if m_optimizationForQuality then
				frq := 1 / (MFPTimerFrq / SE2TC / (FrqAY / 8))
			else
				frq := 1 / (MFPTimerFrq / SE2TC / SampleRate);
			case AtariSE2TP of
				0: AtariSE2Channel := 0;
				1: AtariTimerPeriod2 := frq * 4;
				2: AtariTimerPeriod2 := frq * 10;
				3: AtariTimerPeriod2 := frq * 16;
				4: AtariTimerPeriod2 := frq * 50;
				5: AtariTimerPeriod2 := frq * 64;
				6: AtariTimerPeriod2 := frq * 100;
				7: AtariTimerPeriod2 := frq * 200
			end
		end
		else
			AtariSE2Channel := 0;
		if AtariTimerCounter2 >= AtariTimerPeriod2 then
			AtariTimerCounter2 := 0
	end
	else
		AtariTimerCounter2 := 0;

	inc(Global_Tick_Counter);
	inc(Position_In_VTX, 16)
end;

procedure CYmEngine.synthesizerYM6;
begin
	if not IntFlag then
	begin
		inc(Number_Of_Tiks.re, YM6Tiks);
		if Number_Of_Tiks.hi = 0 then exit
	end
	else
		IntFlag := False;
	Synthesizer;
end;


// Initialisation for all types
//
procedure CYmEngine.initForAllTypes;
var
	i: word;
begin
	case FileType of
		YM3File..YM3bFile: Position_In_VTX := 0;
		YM5File, YM6File:
		begin
			Position_In_VTX := 0;
			AtariSE1Channel := 0;
			AtariSE2Channel := 0;
			AtariTimerCounter1 := 0;
			AtariTimerCounter2 := 0;

			if m_optimizationForQuality then
				YM6CurTik := YM6TiksOnInt
			else
				YM6CurTik := YM6SamTiksOnInt;

			AtariV1 := 0;
			AtariV2 := 0;
			YM6SinusPos1 := 0;
			YM6SinusPos2 := 0
		end;
	end;

	for i := 0 to 12 do RegisterAY.Index[i] := 0;
	SetEnvelopeRegister(0);
	SetMixerRegister(0);
	SetAmplA(0);
	SetAmplB(0);
	SetAmplC(0);
	First_Period := False;
	Ampl := 0;
	Current_RegisterAY := 0;
	if not m_isPlaying then
	begin
		resetAYChipEmulation;
		Global_Tick_Counter := 0;
		ProgrMax := round(Time_Rasch / 1000 * SampleRate * (SampleBit div 8) * NumOfChan);
		Real_End := False;
		FileOutDataCounter := 0;
	end
end;

procedure CYmEngine.resetAYChipEmulation;
begin
	IntFlag := False;
	Number_Of_Tiks.Re := 0;
	Current_Tik := 0;
	PBuffer := Buffer1;
	Envelope_Counter.Re := 0;
	Ton_Counter_A.Re := 0;
	Ton_Counter_B.Re := 0;
	Ton_Counter_C.Re := 0;
	Noise_Counter.Re := 0;
	Ton_A := 0;
	Ton_B := 0;
	Ton_C := 0;
	Left_Chan := 0;
	Right_Chan := 0;
	Tick_Counter := 0;
	Tik.Re := Delay_In_Tiks;
	Noise.Seed := $ffff;
	Reseted := False
end;

function CYmEngine.revers(a: integer; b: integer): dword;
var
	i, s: integer;
begin
	s := 0;

	for i := 0 to pred(b) do
		s := s or (((a and (255 shl (8 * i))) shr (8 * i)) shl (8 * (b - i - 1)));

	revers := s;
end;

// WaveOut error message
//
procedure CYmEngine.raiseWaveOutError(Text: PChar; Title: string);
begin
	raise Exception.Create('[' + Title + ']' + #13#10 + Text);
end;

procedure CYmEngine.calculateLevelTables;
var
	i, b, l, r: integer;
	Index_A, Index_B, Index_C: integer;
	k: real;
begin
	Index_A := Index_AL;
	Index_B := Index_BL;
	Index_C := Index_CL;
	l := Index_A + Index_B + Index_C;
	r := Index_AR + Index_BR + Index_CR;
	if NumOfChan = 2 then
	begin
		if l < r then
			l := r;
	end
	else
	begin
		l := l + r;
		inc(Index_A, Index_AR);
		inc(Index_B, Index_BR);
		inc(Index_C, Index_CR)
	end;
	if l = 0 then inc(l);
	if SampleBit = 8 then r := 127
	else
		r := 32767;
	l := 255 * r div l;
	case Emulating_Chip of
		YM_Chip: for i := 0 to 31 do
			begin
				b := round(Index_A / 255 * Amplitudes_YM[i]);
				Level_AL[i] := round(b / 65535 * l);
				b := round(Index_AR / 255 * Amplitudes_YM[i]);
				Level_AR[i] := round(b / 65535 * l);
				b := round(Index_B / 255 * Amplitudes_YM[i]);
				Level_BL[i] := round(b / 65535 * l);
				b := round(Index_BR / 255 * Amplitudes_YM[i]);
				Level_BR[i] := round(b / 65535 * l);
				b := round(Index_C / 255 * Amplitudes_YM[i]);
				Level_CL[i] := round(b / 65535 * l);
				b := round(Index_CR / 255 * Amplitudes_YM[i]);
				Level_CR[i] := round(b / 65535 * l);
			end;
	end;
	k := (exp(m_preAmp * ln(2) / PreAmpMax) - 1);
	for i := 0 to 31 do
	begin
		Level_AL[i] := round(Level_AL[i] * k);
		Level_AR[i] := round(Level_AR[i] * k);
		Level_BL[i] := round(Level_BL[i] * k);
		Level_BR[i] := round(Level_BR[i] * k);
		Level_CL[i] := round(Level_CL[i] * k);
		Level_CR[i] := round(Level_CR[i] * k);
	end;
end;

procedure CYmEngine.setChipFrq(Fr: integer);
begin
	if (Fr >= 1000000) and (Fr <= 3000000) then
	begin
		FrqAY := Fr;
		if MFPTimerMode = 0 then
			MFPTimerFrq := round(FrqAY * 1.25);
		Delay_In_Tiks := round(8192 / SampleRate * FrqAY);
		//  FrqAyByFrqZ80 := round(FrqAY/FrqZ80/8 * 4294967296);
		Tik.Re := Delay_In_Tiks;
		Tiks_In_Interrupt := round(FrqAY / (Interrupt_Freq / 1000 * 8));
		YM6TiksOnInt := FrqAY / (Interrupt_Freq / 1000 * 8);
	end
end;

// Sets the quality flag (0 -> Performance 1 -> Quality)
//
procedure CYmEngine.setOptimizationFlag(Q: boolean);
begin
	if m_optimizationForQuality = Q then exit;
	m_optimizationForQuality := Q;
	if Q then
	begin
		AtariTimerPeriod1 := AtariTimerPeriod1 / SampleRate * (FrqAY div 8);
		AtariTimerPeriod2 := AtariTimerPeriod2 / SampleRate * (FrqAY div 8);
		AtariTimerCounter1 := AtariTimerCounter1 / SampleRate * (FrqAY div 8);
		AtariTimerCounter2 := AtariTimerCounter2 / SampleRate * (FrqAY div 8);
		YM6CurTik := YM6CurTik / SampleRate * (FrqAY div 8);
		Current_Tik := round(Current_Tik / SampleRate * (FrqAY div 8));
		Number_Of_Tiks.Re := round(Number_Of_Tiks.Re / SampleRate * (FrqAY div 8));
		if NumOfChan = 2 then
		begin
			if SampleBit = 8 then
			begin
				Synthesizer := Synthesizer_Stereo8;
			end
			else
			begin
				Synthesizer := Synthesizer_Stereo16;
			end
		end
		else if SampleBit = 8 then
		begin
			Synthesizer := Synthesizer_Mono8;
		end
		else
		begin
			Synthesizer := Synthesizer_Mono16;
		end
	end
	else
	begin
		AtariTimerPeriod1 := AtariTimerPeriod1 / (FrqAY div 8) * SampleRate;
		AtariTimerPeriod2 := AtariTimerPeriod2 / (FrqAY div 8) * SampleRate;
		AtariTimerCounter1 := AtariTimerCounter1 / (FrqAY div 8) * SampleRate;
		AtariTimerCounter2 := AtariTimerCounter2 / (FrqAY div 8) * SampleRate;
		YM6CurTik := YM6CurTik / (FrqAY div 8) * SampleRate;
		Current_Tik := round(Current_Tik / (FrqAY div 8) * SampleRate);
		Number_Of_Tiks.Re := round(Number_Of_Tiks.Re / (FrqAY div 8) * SampleRate);
		if NumOfChan = 2 then
		begin
			if SampleBit = 8 then
			begin
				Synthesizer := Synthesizer_Stereo8_P;
			end
			else
			begin
				Synthesizer := Synthesizer_Stereo16_P;
			end
		end
		else if SampleBit = 8 then
		begin
			Synthesizer := Synthesizer_Mono8_P;
		end
		else
		begin
			Synthesizer := Synthesizer_Mono16_P;
		end
	end;
end;

// Gets all ym information from decrunched buffer
//
procedure CYmEngine.getYmInformation;
var
	i, i1, i2, k, k2: integer;
	j: longword;
	sh: shortint;
	b: byte;
begin
	Move(m_ymDecruncher.DecrunchedData^, i, 4);

	case longword(i) of

		// Ym 3b file
		$62334d59:
		begin
			FileType := YM3bFile;
			m_ymSong.FileType := YM3bFile;
			Number_Of_Tiks_In_VTX := (m_ymDecruncher.Header.UCompSize - 8) div 14;
			Move(m_ymDecruncher.DecrunchedData^[m_ymDecruncher.Header.UCompSize - 4],
				VBL_For_Loop, 4);
			if m_ymSong.Loop_VBL = -1 then
				m_ymSong.Loop_VBL := VBL_For_Loop
			else
				VBL_For_Loop := m_ymSong.Loop_VBL;

			VTX_Offset := 4;
		end;

		// Ym 3 file
		$21334d59, $21324d59:
		begin
			FileType := YM3File;
			Number_Of_Tiks_In_VTX := (m_ymDecruncher.Header.UCompSize - 4) div 14;
			VTX_Offset := 4;
		end;

		// Ym 5 file
		$21354d59, $21364d59:
		begin
			if longword(i) = $21354d59 then
			begin
				m_ymSong.FileType := YM5File;
				FileType := YM5File
			end
			else
			begin
				m_ymSong.FileType := YM6File;
				FileType := YM6File
			end;
			Move(m_ymDecruncher.DecrunchedData^, YM5Header, 34);
			YM5Header.loop := revers(YM5Header.loop, SizeOf(YM5Header.loop));
			VBL_For_Loop := YM5Header.loop;

			if m_ymSong.Loop_VBL = -1 then
				m_ymSong.Loop_VBL := VBL_For_Loop
			else
				VBL_For_Loop := m_ymSong.Loop_VBL;

			YM5Header.Num_of_tiks := revers(YM5Header.Num_of_tiks,
				SizeOf(YM5Header.Num_of_tiks));
			Number_Of_Tiks_In_VTX := YM5Header.Num_of_tiks;
			YM5Header.InterFrq := revers(YM5Header.InterFrq, SizeOf(YM5Header.InterFrq));
			YM5Header.ChipFrq := revers(YM5Header.ChipFrq, SizeOf(YM5Header.ChipFrq));

			with m_ymSong do
			begin
				if SoundChip_Frq = -2 then
					SoundChip_Frq := YM5Header.ChipFrq;
				if Player_Frq = -2 then
					Player_Frq := YM5Header.InterFrq * 1000;
			end;

			VTX_Offset := 34 + revers(YM5Header.Add_Size, Sizeof(YM5Header.Add_Size));
			i := revers(YM5Header.Num_of_Dig, SizeOf(YM5Header.Num_of_Dig));
			SetLength(DDrumSamples, i);
			i1 := 0;
			while i > 0 do
			begin
				j := DWord(m_ymDecruncher.DecrunchedData^[VTX_Offset]) shl 24;
				inc(VTX_Offset);
				inc(j, DWord(m_ymDecruncher.DecrunchedData^[VTX_Offset]) * 65536);
				inc(VTX_Offset);
				inc(j, Word(m_ymDecruncher.DecrunchedData^[VTX_Offset]) * 256);
				inc(VTX_Offset);
				inc(j, m_ymDecruncher.DecrunchedData^[VTX_Offset]);
				inc(VTX_Offset);
				DDrumSamples[i1].Length := j;
				DDrumSamples[i1].Buf := @m_ymDecruncher.DecrunchedData^[VTX_Offset];
				inc(VTX_Offset, j);
				inc(i1);
				dec(i)
			end;

			if YM5Header.Song_Attr and $06000000 = $02000000 then
				for i := 0 to Length(DDrumSamples) - 1 do
					for i1 := 0 to DDrumSamples[i].Length - 1 do
					begin
						sh := shortint(DDrumSamples[i].Buf[i1]);
						k2 := round((sh + 128) / 255 * 65535);
						k := 1;

						while (k < 32) and (Amplitudes_YM[k] < k2) do
							inc(k, 2);

						if k > 1 then
							if Amplitudes_YM[k] - k2 > k2 - Amplitudes_YM[k - 2] then
								dec(k, 2);

						DDrumSamples[i].Buf[i1] := k div 2
					end
			else if YM5Header.Song_Attr and $06000000 = 0 then
				for i := 0 to Length(DDrumSamples) - 1 do
					for i1 := 0 to DDrumSamples[i].Length - 1 do
					begin
						b := DDrumSamples[i].Buf[i1];
						k2 := round(b / 255 * 65535);
						k := 1;

						while (k < 32) and (Amplitudes_YM[k] < k2) do
							inc(k, 2);

						if k > 1 then
							if Amplitudes_YM[k] - k2 > k2 - Amplitudes_YM[k - 2] then
								dec(k, 2);

						DDrumSamples[i].Buf[i1] := k div 2
					end;

			if m_ymSong.Mod_Name = '' then
				i1 := 0
			else
				i1 := 1;

			// Get the Song name
			while m_ymDecruncher.DecrunchedData^[VTX_Offset] <> 0 do
			begin
				if i1 = 0 then
					m_ymSong.Mod_Name := m_ymSong.Mod_Name +
						Char(m_ymDecruncher.DecrunchedData^[VTX_Offset]);
				inc(VTX_Offset);
			end;

			// Get the author
			if m_ymSong.Author = '' then
				i2 := 0
			else
				i2 := 1;

			repeat
				inc(VTX_Offset);
				if (i2 = 0) and (m_ymDecruncher.DecrunchedData^[VTX_Offset] <> 0) then
					m_ymSong.Author := m_ymSong.Author +
						Char(m_ymDecruncher.DecrunchedData^[VTX_Offset]);
			until (m_ymDecruncher.DecrunchedData^[VTX_Offset] = 0);

			// Get the comment
			if m_ymSong.Comment_String = '' then
				i := 0
			else
				i := 1;

			repeat
				inc(VTX_Offset);
				if (i = 0) and (m_ymDecruncher.DecrunchedData^[VTX_Offset] <> 0) then
					m_ymSong.Comment_String := m_ymSong.Comment_String +
						Char(m_ymDecruncher.DecrunchedData^[VTX_Offset]);
			until m_ymDecruncher.DecrunchedData^[VTX_Offset] = 0;
			inc(VTX_Offset);
			if i = 0 then
				m_ymSong.Comment_String := Trim(m_ymSong.Comment_String);
		end;
		else
		begin
			m_ymSong.FileType := Unknown;
			FileType := Unknown;
		end
	end;

	if FileType = Unknown then
		exit;

	case FileType of
		YM3File..YM3BFile: MakeBuffer := makeBufferVTX;
		YM5File: MakeBuffer := makeBufferYM5;
		YM6File: MakeBuffer := makeBufferYM6;
	end;

	Emulating_Chip := YM_Chip;

   with m_ymSong do
	begin
		setChipFrq(FrqSTDef);
		calculateLevelTables;

		i := 0;
		if Loop_VBL < 0 then
		begin
			getLength;
			i := 1;
			Loop_VBL := VBL_For_loop;
		end
		else
			VBL_For_loop := Loop_VBL;

		if Time_Length = 0 then
		begin
			if i = 0 then
			begin
				getLength;
				VBL_For_loop := Loop_VBL;
			end;

			Time_Length := Global_Tick_Max;
		end
		else
		begin
			Time_Rasch := round(Time_Length / Interrupt_Freq * 1000000);
			Global_Tick_Max := Time_Length
		end;
	end;
end;

//------------------------------------------------------------------------------
// Property accessors
//------------------------------------------------------------------------------

// Gets the output sample rate
//
function CYmEngine.getSampleRate: TYmSampleFrequency;
begin
   case SampleRate of
   	44100 : result := freq44100;
   end;
end;

// Sets the output sample rate
//
procedure CYmEngine.setSampleRate(theRate: TYmSampleFrequency);
var
	sr: integer;
begin
	sr := 44100;

	case theRate of
		freq11025: sr := 11025;
		freq22050: sr := 22050;
		freq44100: sr := 44100;
	end;

	SampleRate := sr;
	Delay_In_Tiks := round(8192 / SampleRate * FrqAY);
	YM6SamTiksOnInt := SampleRate / Interrupt_Freq * 1000;
	Sample_Tiks_in_Interrupt := Round(SampleRate / Interrupt_Freq * 1000);
end;

// Get the mono / stereo mode
//
function CYmEngine.getOutputChannels: TYmSampleChannels;
begin
	case NumOfChan of
   	1: result := channelsMono;
       2: result := channelsStereo;
   end;
end;

// Sets the mono / stereo mode
//
procedure CYmEngine.setOutputChannels(theChannels: TYmSampleChannels);
var
	nc: integer;
begin
	nc := 2;

	case theChannels of
		channelsMono: nc := 1;
		channelsStereo: nc := 2;
	end;

	NumOfChan := nc;
	MaxSamples := MaxBuffLen div longword(nc);

	// With Optimization
	if m_optimizationForQuality then
	begin
		if nc = 2 then
		begin
			if SampleBit = 8 then
				Synthesizer := Synthesizer_Stereo8
			else
				Synthesizer := Synthesizer_Stereo16;
		end
		else if SampleBit = 8 then
			Synthesizer := Synthesizer_Mono8
		else
			Synthesizer := Synthesizer_Mono16;
	end
	else
	begin
		// Without optimization
		if nc = 2 then
		begin
			if SampleBit = 8 then
				Synthesizer := Synthesizer_Stereo8_P
			else
				Synthesizer := Synthesizer_Stereo16_P;
		end
		else if SampleBit = 8 then
			Synthesizer := Synthesizer_Mono8_P
		else
			Synthesizer := Synthesizer_Mono16_P;
	end;

	calculateLevelTables;
end;

// Get the output sample bits
//
function CYmEngine.getOutputBits: TYmSampleBits;
begin
	case SampleBit of
   	8: result := bits8;
       16: result := bits16;
   end;
end;

// Sets the output sample bits
//
procedure CYmEngine.setOutputBits(theBits: TYmSampleBits);
var
	sb: integer;
begin
	sb := 16;

	case theBits of
		bits8: sb := 8;
		bits16: sb := 16;
	end;

	SampleBit := sb;
	MaxBuffLen := 65280 shr (sb div 16);
	MaxSamples := MaxBuffLen div longword(NumOfChan);

	// with optimization
	if m_optimizationForQuality then
	begin
		if NumOfChan = 2 then
		begin
			if sb = 8 then
				Synthesizer := Synthesizer_Stereo8
			else
				Synthesizer := Synthesizer_Stereo16;
		end
		else if sb = 8 then
			Synthesizer := Synthesizer_Mono8
		else
			Synthesizer := Synthesizer_Mono16;
	end
	else
	// without optimization
	begin
		if NumOfChan = 2 then
		begin
			if sb = 8 then
				Synthesizer := Synthesizer_Stereo8_P
			else
				Synthesizer := Synthesizer_Stereo16_P;
		end
		else if sb = 8 then
			Synthesizer := Synthesizer_Mono8_P
		else
			Synthesizer := Synthesizer_Mono16_P;
	end;

	calculateLevelTables;
end;


// Sets the channel mode
//
procedure CYmEngine.setMixerMode(theMixerMode: TYmMixerMode);
begin
	m_channelMode := theMixerMode;
	case theMixerMode of
		modeMono:
		begin
			Index_AL := 255;
			Index_AR := 255;
			Index_BL := 255;
			Index_BR := 255;
			Index_CL := 255;
			Index_CR := 255;
		end;
		modeABC:
		begin
			Index_AL := 255;
			Index_AR := 13;
			Index_BL := 170;
			Index_BR := 170;
			Index_CL := 13;
			Index_CR := 255;
		end;
		modeACB:
		begin
			Index_AL := 255;
			Index_AR := 13;
			Index_BL := 13;
			Index_BR := 255;
			Index_CL := 170;
			Index_CR := 170;
		end;
		modeBAC:
		begin
			Index_AL := 170;
			Index_AR := 170;
			Index_BL := 255;
			Index_BR := 13;
			Index_CL := 13;
			Index_CR := 255;
		end;
		modeBCA:
		begin
			Index_AL := 13;
			Index_AR := 255;
			Index_BL := 255;
			Index_BR := 13;
			Index_CL := 170;
			Index_CR := 170;
		end;
		modeCAB:
		begin
			Index_AL := 170;
			Index_AR := 170;
			Index_BL := 13;
			Index_BR := 255;
			Index_CL := 255;
			Index_CR := 13;
		end;
		modeCBA:
		begin
			Index_AL := 13;
			Index_AR := 255;
			Index_BL := 170;
			Index_BR := 170;
			Index_CL := 255;
			Index_CR := 13;
		end;
	end;
	calculateLevelTables;
end;

// Get the mixer channels
//
function CYmEngine.getMixerMode: TYmMixerMode;
begin
	result := m_channelMode;
end;

// Sets the optimization mode
//
procedure CYmEngine.setOptimization(theOptimization: TYmOptimizationMode);
begin
	case theOptimization of
   	forPerformance: setOptimizationFlag(false);
       forQuality: setOptimizationFlag(true);
   end;
end;

// Gets the optimization mode
//
function CYmEngine.getOptimization: TYmOptimizationMode;
begin
   case m_optimizationForQuality of
		false: result := forPerformance;
       true: result := forQuality;
   end;
end;

// Set the Preamp value
//
procedure CYmEngine.setPreampVolume(theVolume: byte);
begin
	m_preAmp := theVolume;
	calculateLevelTables
end;

// Sets the play in loop flag
//
procedure CYmEngine.setPlayInLoop(theValue: boolean);
begin
	m_playInLoop := theValue;
end;


